/**************************************************************
*                                                             *
*  EasyBMP Cross-Platform Windows Bitmap Library Code Sample  * 
*                                                             *
*  Author: Paul Macklin                                       *
*   email: macklin01@users.sourceforge.net                    *
*     WWW: http://easybmp.sourceforge.net                     *
*                                                             *
*    file: Steganography2.cpp                                 *
*    date: 2-9-2005                                           *
* version: 0.59                                               *
*                                                             *
*   License: GPL v 2                                          *
* Copyright: 2005 by the EasyBMP Project                      *
*                                                             *
* description: Code sample for EasyBMP library that embeds a  *
*              text file into a 24-bit BMP file               *
*                                                             *
* algorithm: Regard a byte T as an unsigned character.        *
*            Represent it as T=R1*64+G1*16+R2*4+G2            *
*                 where the Ri,Gi range from 0 to 3.          *
*                                                             *
*            Let (r1,g1,b1) be the original greyscale pixel   *
*                 color. Then overwrite the pixel with        *
*                (r1-R1,g1-G1,b1). Next, let (r2,g2,b2) be    *
*                 the next pixel, and overwrite it with       *
*                (r2-R2,g2-G2,b2). Notice that b1 and b2 are  *
*                 are unchanged, so the original greyscale    *
*                 values are still here.                      *
*            Use the b channels to extract the message: If    *
*                 (rr1,gg1,bb1) and (rr2,gg2,bb2) are the     *
*                  pixles, then you can get R1 = bb1-rr1      *
*                  G1 = bb1-gg1 , and so forth.               *
*                                                             *
**************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <iostream.h>
#include <fstream.h>
#include <math.h>

#include <time.h>

#include "EasyBMP.h"

int main( int argc, char* argv[] )
{
 char* DATE = "May 18";
 char* YEAR = "2005";
 char* AUTHOR = "Paul Macklin";
 char* LICENSE = "GPL v. 2";
 char* GROUP = "the EasyBMP Project";
 char* CONTACT = "http://easybmp.sourceforge.net";
 
 // determine if the user asked for help
 bool HelpMode = false;
 if( argc > 1 )
 {
  char HelpTemp [1025];
  strcpy( HelpTemp , argv[1] );
  if( HelpTemp[0] == '-' && HelpTemp[1] == '-' && 
      HelpTemp[2] == 'h' && HelpTemp[3] == 'e' &&
      HelpTemp[4] == 'l' && HelpTemp[5] == 'p' )
  { HelpMode = true; }
 }
 // if the user asked for help or is confused, spit out help info
 if( argc == 1 || HelpMode )
 {
  cout << "\nEmbedText usage: \n\n"
       << "This embeds a text file into a graphic file, and saves the result to \n"
       << "a new file (-e for encode):\n\n"
       << "          Stego2 -e <input textfile> <input bitmap> <output bitmap> \n\n"
       << "This extracts a hidden text message and outputs the result \n"
       << "(-d for decode):\n\n"
       << "          Stego2 -d <input bitmap> <output textfile>\n\n"
       << "This gives this help information:\n\n"
       << "          Stego2 --help\n\n"
       << " Created on " << DATE << ", " << YEAR << " by " << AUTHOR << ".\n"
       << " Uses the EasyBMP library, Version " << _EasyBMP_Version_ << ".\n"
       << " Licensed under " << LICENSE << " by " << GROUP << ".\n"
       << " Copyright (c) " << YEAR << " " << GROUP << "\n"
       << " Contact: " << CONTACT << "\n\n";
  return 1;
 }
 // spit out copyright / license information
 cout  << "\n Created on " << DATE << ", " << YEAR << " by " << AUTHOR << ".\n"
       << " Uses the EasyBMP library, Version " << _EasyBMP_Version_ << ".\n"
       << " Licensed under " << LICENSE << " by " << GROUP << ".\n"
       << " Copyright (c) " << YEAR << " " << GROUP << "\n"
       << " Contact: " << CONTACT << "\n\n";
       
 // check to see if it's time to encode or decode

 bool EncodeMode = false;
 bool DecodeMode = false;
 char FirstArgumentTemp [1025];
 strcpy( FirstArgumentTemp , argv[1] );
 if( FirstArgumentTemp[0] == '-' && FirstArgumentTemp[1] == 'e'  )
 { EncodeMode = true; } 
 if( FirstArgumentTemp[0] == '-' && FirstArgumentTemp[1] == 'd'  )
 { DecodeMode = true; } 
 
 if( EncodeMode )
 {
  // open the input BMP file
  BMP Image;
  Image.ReadFromFile( argv[3] );
  Image.SetBitDepth( 24 );
  int MaxNumberOfCharacters = Image.TellWidth() * Image.TellHeight() ;
  int k=1;
  
  // open the input text file
  fstream InputText( argv[2] , ios::in );
  
  // determine the noise factor
  
  int NoiseFactor;
  char szNoiseFactorInput [2049];
  cout << "\nEnter the desired noise:embedded info factor (0 allowed): ";
  cin >> szNoiseFactorInput;
  NoiseFactor = (int) strtod( szNoiseFactorInput , NULL );
  NoiseFactor *= 3;
  
  // convert to greyscale, and add some noise
  
  srand( time(NULL) );
  rand();
  rand();
  
  int i=0; 
  int j=0;
  for( i=0 ; i<Image.TellWidth() ; i++)
  {
   for( j=0 ; j<Image.TellHeight() ;j++)
   {
    double Total = pow( (double) Image(i,j)->Red , 2.0 )
                 + pow( (double) Image(i,j)->Green , 2.0 )
                 + pow( (double) Image(i,j)->Blue , 2.0 );
    Total /= 3.0;
    Total = sqrt( Total );
    int NewTotal = (int) Total;
    
    // add random noise 
    if( NoiseFactor )
    {
     int Random = (int) ( ( 2.0 * NoiseFactor * rand() ) / ( RAND_MAX +0.0 ) ) - NoiseFactor;
     NewTotal += Random;
    }
    
    if( NewTotal < 0 )
    { NewTotal = 0; }
    if( NewTotal > 255 )
    { NewTotal = 255; }
    
    Image(i,j)->Red = (BYTE) NewTotal;
    Image(i,j)->Green = (BYTE) NewTotal;
    Image(i,j)->Blue = (BYTE) NewTotal;
   }
  }
  
  // encode each pixel
  i=0; 
  j=0;
  
  bool Stop = false;
  
  while( !InputText.eof() && 2*k <= MaxNumberOfCharacters && !Stop )
  {
   char InputCharacter = InputText.get();
   unsigned int CharacterInt = (unsigned int) InputCharacter;
   
   // write the character as 
   // RedChannel_One*4^3 + GreenChannel_One*4^2 
   // + RedChannel_Two*4 + GreenChannel_Two
   
   unsigned int RedChannel_One = CharacterInt / 64;
   unsigned int Temp = CharacterInt - 64*RedChannel_One; 
   unsigned int GreenChannel_One = Temp / 16;
   Temp = Temp - 16*GreenChannel_One;
   
   if( CharacterInt < 255 )
   {
    // make sure that you don't get overflow upon the impending 
    // subtraction.   
   
    if( RedChannel_One > (unsigned int) Image(i,j)->Red )
    {
     Image(i,j)->Red = (BYTE) RedChannel_One;
     Image(i,j)->Green = (BYTE) RedChannel_One;
     Image(i,j)->Blue = (BYTE) RedChannel_One;
    }
    if( GreenChannel_One > (unsigned int) Image(i,j)->Green )
    {
     Image(i,j)->Red = (BYTE) GreenChannel_One;
     Image(i,j)->Green = (BYTE) GreenChannel_One;
     Image(i,j)->Blue = (BYTE) GreenChannel_One;
    }
    
    // do the actual embedding.
   
    Image(i,j)->Red -= (BYTE) RedChannel_One;
    Image(i,j)->Green -= (BYTE) GreenChannel_One;
    
    // increment the current pixel
    
    i++;
    if( i >= Image.TellWidth() )
    { i = 0; j++; }
    
    unsigned int RedChannel_Two = Temp / 4;
    Temp = Temp - 4*RedChannel_Two; 
    unsigned int GreenChannel_Two = Temp;
   
    // make sure that you don't get overflow upon the impending 
    // subtraction.   
   
    if( RedChannel_Two > (unsigned int) Image(i,j)->Red )
    {
     Image(i,j)->Red = (BYTE) RedChannel_Two;
     Image(i,j)->Green = (BYTE) RedChannel_Two;
     Image(i,j)->Blue = (BYTE) RedChannel_Two;
    }
    if( GreenChannel_Two > (unsigned int) Image(i,j)->Green )
    {
     Image(i,j)->Red = (BYTE) GreenChannel_Two;
     Image(i,j)->Green = (BYTE) GreenChannel_Two;
     Image(i,j)->Blue = (BYTE) GreenChannel_Two;
    }
    
    // do the actual embedding.
   
    Image(i,j)->Red -= (BYTE) RedChannel_Two;
    Image(i,j)->Green -= (BYTE) GreenChannel_Two;   
    
    // increment the current pixel
    
    i++;
    if( i >= Image.TellWidth() )
    { i = 0; j++; }
    
   }
   else
   {
    // nonsense character. It must be an EOF, so it's time to stop
   
    Stop = true; 
   }
   
   
   k++;
  }
  InputText.close();
  
  Image.WriteToFile( argv[4] );
 
  return 0;
 }
 
 if( DecodeMode )
 {
  if( GetBitmapColorDepth( argv[2] ) != 24 )
  { cout << "\n **Error** Use a 24-bit BMP file! \n\n"; return 1; } 
 
  // open the bmp file
  BMP Image;
  Image.ReadFromFile( argv[2] );
  
  // prep the output stream
  
  fstream OutputText( argv[3] , ios::out );
  int k = 1;
  int MaxCharacters = Image.TellWidth() * Image.TellHeight();
  
  int i=0;
  int j=0;
  
  bool Stop = false;
  int BlankCount = 0;
  
  while( 2*k <= MaxCharacters && !Stop )
  {
   unsigned int CharacterInt = (unsigned int) (Image(i,j)->Blue - Image(i,j)->Red)*64 
                             + (unsigned int) (Image(i,j)->Blue - Image(i,j)->Green)*16;
			     
   i++;
   if( i >= Image.TellWidth() )
   { i = 0; j++; }		
   
   CharacterInt += (unsigned int) (Image(i,j)->Blue - Image(i,j)->Red)*4 
                 + (unsigned int) (Image(i,j)->Blue - Image(i,j)->Green);	     
		 
   i++;
   if( i >= Image.TellWidth() )
   { i = 0; j++; }		 
 			     
   char Character = (char) CharacterInt;  
   
   if( CharacterInt != 0 )
   {
    OutputText.put( Character );
   }
   else
   {
    Stop = true; 
   }
  

   
   k++;  
  }
  OutputText.close(); 
 
  return 0;
 }
 
 return 0;
}
